from typing import Iterable, Any, Tuple

import numpy as np
import scipy.spatial

try:
    from scipy.spatial import Delaunay as spDelaunay
except Exception:  # pragma: no cover
    from scipy.spatial.qhull import Delaunay as spDelaunay

from .utils.topology import unique_topo_data
from .utils.tri import edges_tri

from .config import __hasvtk__, __haspyvista__, __hasmatplotlib__

if __hasvtk__:
    from vtk import vtkIdList

if __haspyvista__:
    import pyvista as pv

if __hasmatplotlib__:
    import matplotlib.tri as tri


__all__ = ["triangulate"]


def triangulate(
    *args,
    points: np.ndarray = None,
    size: tuple = None,
    shape: tuple = None,
    origo: tuple = None,
    backend: str = "mpl",
    random: bool = False,
    triangles: Iterable = None,
    triobj=None,
    return_lines: bool = False,
    **kwargs,
):
    """
    Crates a triangulation using different backends.

    Parameters
    ----------
    points: numpy.ndarray, Optional
        A 2d array of coordinates of a flat surface. This or `shape`
        must be provided.
    size: tuple, Optional
        A 2-tuple, containg side lengths of a rectangular domain.
        Should be provided alongside `shape`. Must be provided if
        points is None.
    shape: tuple or int, Optional
        A 2-tuple, describing subdivisions along coordinate directions,
        or the number of base points used for triangulation.
        Should be provided alongside `size`.
    origo: numpy.ndarray, Optional
        1d float array, specifying the origo of the mesh.
    backend: str, Optional
        The backend to use. It can be 'mpl' (matplotlib), 'pv' (pyvista),
        or 'scipy'. Default is 'mpl'.
    random: bool, Optional
        If `True`, and points are provided by `shape`, it creates random
        points in the region defined by `size`. Default is False.
    triangles: numpy.ndarray
        2d integer array of vertex indices. If both `points` and
        `triangles` are specified, the only thing this function does is to
        create a triangulation object according to the argument `backend`.
    triobj: object
        An object representing a triangulation. It can be
        - a result of a call to matplotlib.tri.Triangulation
        - a Delaunay object from scipy
        - an instance of pyvista.PolyData
        In this case, the function can be used to transform between
        the supported backends.
    return_lines: bool, Optional
        If `True` the function return the unique edges of the
        triangulation and the indices of the edges for each triangle.

    Returns
    -------
    numpy.ndarray
        A 2d float array of coordinates of points.
    numpy.ndarray
        A 2d integer array representing the topology.
    MeshLike
        An object representing the triangulation, according to the specified backend.

    Examples
    --------
    Triangulate a rectangle of size 800x600 with a subdivision of 10x10

    >>> from sigmaepsilon.mesh import triangulate
    >>> coords, topo, triobj = triangulate(size=(800, 600), shape=(10, 10))
    ...

    Triangulate a rectangle of size 800x600 with a number of 100 randomly
    distributed points

    >>> coords, topo, triobj = triangulate(size=(800, 600), shape=100, random=True)
    ...
    """
    if len(args) > 0:
        if _is_triobj(args[0]):
            triobj = args[0]

    if triobj is not None:
        points, triangles = get_triobj_data(triobj, *args, **kwargs)
    else:
        # create points from input
        if points is None:
            assert size is not None, (
                "Either a collection of points, or the size of a "
                "rectangular domain must be provided!"
            )

            if origo is None:
                origo = (0, 0, 0)
            else:
                if len(origo) == 2:
                    origo = origo + (0,)

            if shape is None:
                shape = (3, 3)

            if isinstance(shape, int):
                if random:
                    x = np.hstack(
                        [np.array([0, 1, 1, 0], dtype=float), np.random.rand(shape)]
                    )
                    y = np.hstack(
                        [np.array([0, 0, 1, 1], dtype=float), np.random.rand(shape)]
                    )
                    z = np.zeros(len(x), dtype=float)
                    points = np.c_[
                        x * size[0] - origo[0], y * size[1] - origo[1], z - origo[2]
                    ]
                else:
                    shape = (shape, shape)

            if points is None and isinstance(size, tuple):
                x = np.linspace(-origo[0], size[0] - origo[0], num=shape[0])
                y = np.linspace(-origo[1], size[1] - origo[1], num=shape[1])
                z = np.zeros(len(x), dtype=float) - origo[2]
                xx, yy = np.meshgrid(x, y)
                zz = np.zeros(xx.shape, dtype=xx.dtype)
                # Get the points as a 2D NumPy array (N by 2)
                points = np.c_[xx.reshape(-1), yy.reshape(-1), zz.reshape(-1)]

        # generate triangles
        if triangles is None:
            if backend == "mpl":
                triobj = tri.Triangulation(points[:, 0], points[:, 1])
                triangles = triobj.triangles
            elif backend == "scipy":
                triobj = scipy.spatial.Delaunay(points[:, 0:2])
                triangles = triobj.simplices
            elif backend == "pv":
                if not __haspyvista__ or not __hasvtk__:
                    raise ImportError("PyVista must be installed for this.")
                cloud = pv.PolyData(points)
                triobj = cloud.delaunay_2d()
                nCell = triobj.n_cells
                triangles = np.zeros((nCell, 3), dtype=int)
                for cellID in range(nCell):
                    idlist = vtkIdList()
                    triobj.GetCellPoints(cellID, idlist)
                    n = idlist.GetNumberOfIds()
                    triangles[cellID] = [idlist.GetId(i) for i in range(n)]
        else:
            assert backend == "mpl", (
                "This feature is not yet supported by "
                "other backends, only matplotlib."
            )
            triobj = tri.Triangulation(points[:, 0], points[:, 1], triangles=triangles)

    if return_lines:
        edges, edgeIDs = unique_topo_data(edges_tri(triangles))
        return points, edges, triangles, edgeIDs, triobj

    return points, triangles, triobj


def triobj_to_mpl(triobj, *args, **kwargs) -> tri.Triangulation:
    """
    Transforms a triangulation into a matplotlib.tri.Triangulation
    object.
    """
    assert _is_triobj(triobj)
    if isinstance(triobj, tri.Triangulation):
        return triobj
    else:
        points, triangles = get_triobj_data(triobj, *args, **kwargs)
        kwargs["backend"] = "mpl"
        _, _, triang = triangulate(*args, points=points, triangles=triangles, **kwargs)
        return triang


def get_triobj_data(
    obj: Any = None, *_, trim2d: bool = True, **__
) -> Tuple[np.ndarray]:
    coords, topo = None, None

    if isinstance(obj, spDelaunay):
        coords = obj.points
        topo = obj.simplices
    elif isinstance(obj, tri.Triangulation):
        coords = np.vstack((obj.x, obj.y)).T
        topo = obj.triangles
    else:
        if __haspyvista__ and __hasvtk__:
            if isinstance(obj, pv.PolyData):
                if trim2d:
                    coords = obj.points[:, 0:2]
                else:
                    coords = obj.points
                triang = obj.delaunay_2d()
                nCell = triang.n_cells
                topo = np.zeros((nCell, 3), dtype=np.int32)
                for cellID in range(nCell):
                    idlist = vtkIdList()
                    triang.GetCellPoints(cellID, idlist)
                    n = idlist.GetNumberOfIds()
                    topo[cellID] = [idlist.GetId(i) for i in range(n)]

    if coords is None or topo is None:
        raise RuntimeError("Failed to recognize a valid input.")

    return coords, topo


def _is_triobj(triobj: Any) -> bool:
    try:
        if isinstance(triobj, spDelaunay) or isinstance(triobj, tri.Triangulation):
            return True
        else:
            if __haspyvista__:
                if isinstance(triobj, pv.PolyData):
                    if hasattr(triobj, "delaunay_2d"):
                        return True
        return False
    except Exception:  # pragma: no cover
        return False



# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
from numba import njit, prange

__cache = True


@njit(nogil=True, parallel=True, cache=__cache)
def extrude_T3_TET4(points: ndarray, triangles: ndarray, h: float = 1.0, zres: int = 1):
    nT = triangles.shape[0]
    nP = points.shape[0]
    nE = nT * zres * 3
    nC = nP * (zres + 1)
    coords = np.zeros((nC, 3), dtype=points.dtype)
    topo = np.zeros((nE, 4), dtype=triangles.dtype)
    coords[:nP, :2] = points[:, :2]
    for i in prange(zres):
        coords[nP * (i + 1) : nP * (i + 2), :2] = points[:, :2]
        coords[nP * (i + 1) : nP * (i + 2), 2] = h * (i + 1) / zres
        for j in prange(nT):
            id = i * nT * 3 + j * 3
            i_0, j_0, k_0 = triangles[j] + i * nP
            i_1, j_1, k_1 = triangles[j] + (i + 1) * nP
            #
            topo[id, 0] = i_0
            topo[id, 1] = j_0
            topo[id, 2] = k_0
            topo[id, 3] = k_1
            #
            topo[id + 1, 0] = i_0
            topo[id + 1, 1] = j_0
            topo[id + 1, 2] = k_1
            topo[id + 1, 3] = j_1
            #
            topo[id + 2, 0] = i_0
            topo[id + 2, 1] = j_1
            topo[id + 2, 2] = k_1
            topo[id + 2, 3] = i_1
    return coords, topo


@njit(nogil=True, parallel=True, cache=__cache)
def extrude_T3_W6(points: ndarray, triangles: ndarray, h: float = 1.0, zres: int = 1):
    nT = triangles.shape[0]
    nP = points.shape[0]
    nE = nT * zres
    nC = nP * (zres + 1)
    coords = np.zeros((nC, 3), dtype=points.dtype)
    topo = np.zeros((nE, 6), dtype=triangles.dtype)
    coords[:nP, :2] = points[:, :2]
    for i in prange(zres):
        coords[nP * (i + 1) : nP * (i + 2), :2] = points[:, :2]
        coords[nP * (i + 1) : nP * (i + 2), 2] = h * (i + 1) / zres
        for j in prange(nT):
            id = i * nT + j
            i_0, j_0, k_0 = triangles[j] + i * nP
            i_1, j_1, k_1 = triangles[j] + (i + 1) * nP
            #
            topo[id, 0] = i_0
            topo[id, 1] = j_0
            topo[id, 2] = k_0
            topo[id, 3] = i_1
            topo[id, 4] = j_1
            topo[id, 5] = k_1
    return coords, topo



from typing import Union, Callable, Tuple, Optional
from numbers import Number

import numpy as np
from numpy import ndarray

from .typing import PolyCellProtocol
from .data import PointData, PolyData, TriMesh
from .grid import grid
from .cells import H8, H27, TET4, TET10, T3, W6, W18
from .space import CartesianFrame
from .triang import triangulate
from .utils import cell_centers_bulk
from .utils.topology import detach, H8_to_TET4, H27_to_TET10, TET4_to_TET10, W6_to_W18
from .extrude import extrude_T3_TET4, extrude_T3_W6
from .voxelize import voxelize_cylinder


def circular_helix(
    a: Optional[Union[Number, None]] = None,
    b: Optional[Union[Number, None]] = None,
    *,
    slope: Optional[Union[Number, None]] = None,
    pitch: Optional[Union[Number, None]] = None,
) -> Callable[[Number], Tuple[float, float, float]]:
    """
    Returns the function :math:`f(t) = [a \cdot cos(t), a \cdot sin(t), b \cdot t]`,
    which describes a circular helix of radius a and slope a/b (or pitch 2πb).

    Parameters
    ----------
    a: float, Optional
        Radius of the helix. Default is None.
    b: float, Optional
        Slope of the helix. Default is None.
    slope: float, Optional
        Slope of the helix. Default is None.
    pitch: float, Optional
        Pitch of the helix. Default is None.
    """
    if pitch is not None:
        b = b if b is not None else pitch / 2 / np.pi
    if slope is not None:
        a = a if a is not None else slope * b
        b = b if b is not None else slope / a

    def inner(t: Number) -> Tuple[float, float, float]:
        """
        Evaluates :math:`f(t) = [a \cdot cos(t), a \cdot sin(t), b \cdot t]`.
        """
        return a * np.cos(t), a * np.sin(t), b * t

    return inner


def circular_disk(
    nangles: int,
    nradii: int,
    rmin: float,
    rmax: float,
    frame: Optional[Union[CartesianFrame, None]] = None,
) -> TriMesh:
    """
    Returns the triangulation of a circular disk.

    Parameters
    ----------
    nangles: int
        Number of subdivisions in radial direction.
    nradii: int
        Number of subdivisions in circumferential direction.
    rmin: float
        Inner radius. Can be zero.
    rmax: float
        Outer radius.

    Returns
    -------
    TriMesh

    Examples
    --------
    >>> from sigmaepsilon.mesh.recipes import circular_disk
    >>> mesh = circular_disk(120, 60, 5, 25)
    """
    radii = np.linspace(rmin, rmax, nradii)
    angles = np.linspace(0, 2 * np.pi, nangles, endpoint=False)
    angles = np.repeat(angles[..., np.newaxis], nradii, axis=1)
    angles[:, 1::2] += np.pi / nangles
    x = (radii * np.cos(angles)).flatten()
    y = (radii * np.sin(angles)).flatten()
    nP = len(x)
    points = np.stack([x, y], axis=1)
    *_, triang = triangulate(points=points, backend="mpl")
    # triang = tri.Triangulation(x, y)
    # Mask off unwanted triangles.
    triang.set_mask(
        np.hypot(x[triang.triangles].mean(axis=1), y[triang.triangles].mean(axis=1))
        < rmin
    )
    triangles = triang.get_masked_triangles()
    points = np.stack((triang.x, triang.y, np.zeros(nP)), axis=1)
    points, triangles = detach(points, triangles)
    frame = CartesianFrame(dim=3) if frame is None else frame
    pd = PointData(coords=points, frame=frame)
    cd = T3(topo=triangles, frames=frame)
    return TriMesh(pd, cd)


def cylinder(
    shape: Tuple[Tuple[Number, Number] | Number, Number],
    size: Union[Tuple, float, int],
    *,
    voxelize: bool = False,
    celltype: Optional[Union[PolyCellProtocol, None]] = None,
    frame: Optional[Union[CartesianFrame, None]] = None,
) -> PolyData:
    """
    Returns the coordinates and the topology of cylinder as numpy arrays.

    Parameters
    ----------
    shape: Tuple[Tuple[Number, Number] | Number, Number]
        The parameter is a 2-tuple that describes the dimensions of the
        cylinder. The first element is either a number (radius), or a 2-tuple of
        numbers (inner and outer radii) describing the radius of the cylinder.
        The second element is the height of the cylinder.
    size: Union[tuple, float, int]
        Parameter controlling the resolution of the mesh. Default is None.

        If `voxelize` is ``False``, ``size`` must be a tuple of three
        integers, describing the number of angular, radial, and vertical
        divions in this order.

        If `voxelize` is ``True`` and ``size`` is a ``float``,
        the parameter controls the size of the individual voxels.

        If `voxelize` is ``True`` and ``size`` is an ``int``,
        the parameter controls the size of the individual voxels
        according to :math:`edge \, length = (r_{ext} - r_{int})/shape`.
    voxelize: bool, Optional
        If ``True``, the cylinder gets voxelized to a collection of H8 cells.
        In this case the size of a voxel can be controlled by specifying a
        flaot or an integer as the second parameter ``size``.
        Default is ``False``.
    celltype
        Specifies the celltype to be used.

    Example
    -------
    >>> from sigmaepsilon.mesh.recipes import cylinder
    >>> n_angles = 30  # number of subdivions along the angular direction
    >>> n_radii = 15  # number of subdivisions along the radial direction
    >>> min_radius = 10 # minimum radius of the cylinder
    >>> max_radius = 25 # maximum radius of the cylinder
    >>> n_z = 20 # number of subdivisions along the z direction
    >>> h = 50 # height of the cylinder
    >>> shape = (min_radius, max_radius), h
    >>> size = n_radii, n_angles, n_z
    >>> mesh = cylinder(shape, size, voxelize=False)

    Returns
    -------
    :class:`~sigmaepsilon.mesh.data.polydata.PolyData`
    """
    if celltype is None:
        celltype = H8 if voxelize else TET4
    etype = None

    if isinstance(size, float) or isinstance(size, int):
        size = [size]

    if voxelize:
        etype = "H8"

    radius, h = shape

    if isinstance(radius, int):
        radius = np.array([0, radius])
    elif not isinstance(radius, ndarray):
        radius = np.array(radius)

    etype = celltype.label if etype is None else etype

    if voxelize:
        if isinstance(size[0], int):
            size_ = (radius[1] - radius[0]) / size[0]
        elif isinstance(size[0], float):
            size_ = size[0]
        coords, topo = voxelize_cylinder(radius=radius, height=h, size=size_)
    else:
        if etype == "TET4":
            min_radius, max_radius = radius
            n_radii, n_angles, n_z = size
            mesh = circular_disk(n_angles, n_radii, min_radius, max_radius)
            points, triangles = mesh.coords(), mesh.topology().to_numpy()
            coords, topo = extrude_T3_TET4(points, triangles, h, n_z)
        else:
            raise NotImplementedError("Celltype not supported!")

    frame = CartesianFrame(dim=3) if frame is None else frame
    pd = PointData(coords=coords, frame=frame)
    cd = celltype(topo=topo, frames=frame)
    return PolyData(pd, cd)


def ribbed_plate(
    lx: float,
    ly: float,
    t: float,
    *,
    wx: float = None,
    wy: float = None,
    hx: float = None,
    hy: float = None,
    ex: float = None,
    ey: float = None,
    lmax: float = None,
    order: int = 1,
    tetrahedralize: bool = False,
) -> PolyData:
    """
    Creates a ribbed plate.

    Parameters
    ----------
    lx: float
        The length of the plate along the X axis.
    ly: float
        The length of the plate along the Y axis.
    t: float
        The thickness of a plate.
    wx: float, Optional
        The width of the ribs running in X direction. Must be defined
        alongside `hx`. Default is None.
    hx: float, Optional
        The height of the ribs running in X direction. Must be defined
        alongside `wx`. Default is None.
    ex: float, Optional
        The eccentricity of the ribs running in X direction.
    wy: float, Optional
        The width of the ribs running in Y direction. Must be defined
        alongside `hy`. Default is None.
    hy: float, Optional
        The height of the ribs running in Y direction. Must be defined
        alongside `wy`. Default is None.
    ey: float, Optional
        The eccentricity of the ribs running in Y direction.
    lmax: float, Optional
        Maximum edge length of the cells in the resulting mesh. Default is None.
    order: int, Optional
        Determines the order of the cells used. Allowed values are 1 and 2. If order is
        1, either H8 hexahedra or TET4 tetrahedra are returned. If order is 2, H27
        hexahedra or TET10 tetrahedra are returned.
    tetrahedralize: bool, Optional
        If True, a mesh of 4-noded tetrahedra is returned. Default is False.

    Example
    -------
    >>> from sigmaepsilon.mesh.recipes import ribbed_plate
    >>> mesh = ribbed_plate(lx=5.0, ly=5.0, t=1.0,
    ...                     wx=1.0, hx=2.0, ex=0.05,
    ...                     wy=1.0, hy=2.0, ey=-0.05)

    Returns
    -------
    :class:`~sigmaepsilon.mesh.data.polydata.PolyData`
    """

    def subdivide(bins, lmax):
        _bins = []
        for i in range(len(bins) - 1):
            a = bins[i]
            b = bins[i + 1]
            if (b - a) > lmax:
                ndiv = int(np.ceil((b - a) / lmax))
            else:
                ndiv = 1
            ldiv = (b - a) / ndiv
            for j in range(ndiv):
                _bins.append(a + j * ldiv)
        _bins.append(bins[-1])
        return np.array(_bins)

    xbins, ybins, zbins = [], [], []
    xbins.extend([-lx / 2, 0, lx / 2])
    ybins.extend([-ly / 2, 0, ly / 2])
    zbins.extend([-t / 2, 0, t / 2])

    if wx is not None and hx is not None:
        ex = 0.0 if ex is None else ex
        ybins.extend([-wx / 2, wx / 2])
        if (ex - hx / 2) < (-t / 2):
            zbins.append(ex - hx / 2)
        if (ex + hx / 2) > (t / 2):
            zbins.append(ex + hx / 2)

    if wy is not None and hy is not None:
        ey = 0.0 if ey is None else ey
        xbins.extend([-wy / 2, wy / 2])
        if (ey - hy / 2) < (-t / 2):
            zbins.append(ey - hy / 2)
        if (ey + hy / 2) > (t / 2):
            zbins.append(ey + hy / 2)

    xbins = np.unique(np.sort(xbins))
    ybins = np.unique(np.sort(ybins))
    zbins = np.unique(np.sort(zbins))
    if isinstance(lmax, float):
        xbins = subdivide(xbins, lmax)
        ybins = subdivide(ybins, lmax)
        zbins = subdivide(zbins, lmax)
    bins = xbins, ybins, zbins

    if order == 1:
        coords, topo = grid(bins=bins, eshape="H8")
    elif order == 2:
        coords, topo = grid(bins=bins, eshape="H27")
    else:
        raise ValueError("'order' must be either 1 or 2")

    centers = cell_centers_bulk(coords, topo)
    mask = (centers[:, 2] > (-t / 2)) & (centers[:, 2] < (t / 2))

    if wx is not None and hx is not None:
        m = (centers[:, 1] > (-wx / 2)) & (centers[:, 1] < (wx / 2))
        m = m & (centers[:, 2] > (ex - hx / 2)) & (centers[:, 2] < (ex + hx / 2))
        mask = mask | m

    if wy is not None and hy is not None:
        m = (centers[:, 0] > (-wy / 2)) & (centers[:, 0] < (wy / 2))
        m = m & (centers[:, 2] > (ey - hy / 2)) & (centers[:, 2] < (ey + hy / 2))
        mask = mask | m

    topo = topo[mask, :]

    if tetrahedralize:
        if order == 1:
            coords, topo = H8_to_TET4(coords, topo)
            celltype = TET4
        else:
            coords, topo = H27_to_TET10(coords, topo)
            celltype = TET10
    else:
        celltype = H8 if order == 1 else H27

    coords, topo = detach(coords, topo)
    frame = CartesianFrame(dim=3)
    pd = PointData(coords=coords, frame=frame)
    cd = celltype(topo=topo, frames=frame)
    return PolyData(pd, cd)


def perforated_cube(
    lx: float,
    ly: float,
    lz: float,
    radius: float,
    *,
    nangles: int = None,
    lmax: float = None,
    order: int = 1,
    prismatic: bool = True,
) -> PolyData:
    """
    Returns a cube of side lengths 'lx', 'ly' and 'lz', with a circular hole
    along the 'z' axis.

    Returns
    -------
    :class:`~sigmaepsilon.mesh.data.polydata.PolyData`
    """
    size = (lx, ly)

    if lmax is not None:
        shape = (max([int(lx / lmax), 4]), max([int(ly / lmax), 4]))
    else:
        shape = (4, 4)
    coords, _ = grid(size=size, shape=shape, eshape=(2, 2), centralize=True)

    if lmax is not None:
        where = np.hypot(coords[:, 0], coords[:, 1]) > (radius + lmax)
    else:
        where = np.hypot(coords[:, 0], coords[:, 1]) > (radius * 1.1)
    coords = coords[where]

    if nangles is None:
        if lmax is not None:
            nangles = max(int(2 * np.pi * radius / lmax), 8)
        else:
            nangles = 16

    angles = np.linspace(0, 2 * np.pi, nangles, endpoint=False)
    x_circle = (radius * np.cos(angles)).flatten()
    y_circle = (radius * np.sin(angles)).flatten()
    circle_coords = np.stack([x_circle, y_circle], axis=1)

    coords = np.vstack([coords[:, :2], circle_coords])

    *_, triobj = triangulate(points=coords, backend="mpl")
    triobj.set_mask(
        np.hypot(
            coords[:, 0][triobj.triangles].mean(axis=1),
            coords[:, 1][triobj.triangles].mean(axis=1),
        )
        < radius
    )
    topo = triobj.get_masked_triangles()
    coords = np.stack((triobj.x, triobj.y, np.zeros(coords.shape[0])), axis=1)
    coords, topo = detach(coords, topo)

    if lmax is not None:
        zres = int(lz / lmax)
    else:
        zres = 4

    if prismatic:
        coords, topo = extrude_T3_W6(coords, topo, h=lz, zres=zres)
    else:
        coords, topo = extrude_T3_TET4(coords, topo, h=lz, zres=zres)

    if order == 1:
        celltype = W6 if prismatic else TET4
    elif order == 2:
        if prismatic:
            coords, topo = W6_to_W18(coords, topo)
            celltype = W18
        else:
            coords, topo = TET4_to_TET10(coords, topo)
            celltype = TET10
    else:
        raise ValueError("'order' must be either 1 or 2")

    frame = CartesianFrame(dim=3)
    pd = PointData(coords=coords, frame=frame)
    cd = celltype(topo=topo, frames=frame)
    return PolyData(pd, cd)


def sphere(
    radius: float,
    n_azimuthal_divisions: int,
    n_polar_divisions: Optional[Union[int, None]] = None,
) -> PolyData:
    """
    Generate a triangulated mesh representing a sphere.

    Parameters
    ----------
    radius : float
        Radius of the sphere.
    n_azimuthal_divisions : int
        Number of divisions for azimuthal angle (longitude).
    n_polar_divisions : int, Optional
        Number of divisions for polar angle (latitude). If not provided
        it is equal to 'n_azimuthal_divisions'.

    Returns
    -------
    :class:`~sigmaepsilon.mesh.data.polydata.PolyData`

    Notes
    -----
    This function generates a triangulated mesh of a sphere using separate
    divisions for azimuthal and polar angles. The resulting mesh represents
    a sphere with given radius and resolution.

    Examples
    --------
    >>> from sigmaepsilon.mesh import PolyData
    >>> from sigmaepsilon.mesh.recipes import sphere
    >>> mesh: PolyData = sphere(1.0, 20, 10)
    """

    if not n_polar_divisions:
        n_polar_divisions = n_azimuthal_divisions

    # Create points (vertices)
    phi = np.linspace(0, np.pi, n_polar_divisions)  # Polar angle
    theta = np.linspace(0, 2 * np.pi, n_azimuthal_divisions)  # Azimuthal angle
    phi, theta = np.meshgrid(phi, theta)

    x = radius * np.sin(phi) * np.cos(theta)
    y = radius * np.sin(phi) * np.sin(theta)
    z = radius * np.cos(phi)

    points = np.vstack([x.flatten(), y.flatten(), z.flatten()]).T

    # Create faces (triangles)
    faces = []
    for i in range(n_azimuthal_divisions - 1):
        for j in range(n_polar_divisions - 1):
            p1 = i * n_polar_divisions + j
            p2 = p1 + 1
            p3 = (i + 1) * n_polar_divisions + j
            p4 = p3 + 1

            faces.extend([[p1, p2, p3], [p2, p4, p3]])

    frame = CartesianFrame(dim=3)
    pd = PointData(coords=points, frame=frame)
    cd = T3(topo=np.array(faces))
    mesh = PolyData(pd, cd)
    return mesh



from typing import Tuple, Iterable, Optional, Union
from numbers import Number

import numpy as np
from numpy import ndarray

from sigmaepsilon.math.numint import gauss_points as gp
from .tri import nat_to_loc_tri as n2l_tri, loc_to_nat_tri as l2n_tri
from .tet import nat_to_loc_tet as n2l_tet


class Quadrature:
    def __init__(self, x: Iterable[Number], w: Iterable[Number]):
        self._pos = x
        self._weight = w

    @property
    def pos(self) -> Iterable[Number]:
        return self._pos

    @property
    def weight(self) -> Iterable[Number]:
        return self._weight


# LINES


def Gauss_Legendre_Line_Grid(n: int) -> Tuple[ndarray, ndarray]:
    return gp(n)


#  TRIANGLES

# https://mathsfromnothing.au/triangle-quadrature-rules/?i=1


def _complete_natural_coordinates(nat: ndarray) -> ndarray:
    res = np.zeros((len(nat), 3), dtype=nat.dtype)
    for i in range(len(res)):
        res[i, 2] = 1 - res[i, 0] - res[i, 1]
    return res


def Gauss_Legendre_Tri_1(
    center: Optional[Union[ndarray, None]] = None, natural: Optional[bool] = False
) -> Tuple[ndarray, ndarray]:
    p, w = np.array([[0.0, 0.0]]), np.array([1 / 2])
    if isinstance(center, ndarray):
        p += center
    if natural:
        p = np.array([l2n_tri(x, center=center) for x in p], dtype=float)
    return p, w


def Gauss_Legendre_Tri_3a(
    center: Optional[Union[ndarray, None]] = None, natural: Optional[bool] = False
) -> Tuple[ndarray, ndarray]:
    nat = np.array(
        [
            [2 / 3, 1 / 6, 1 / 6],
            [1 / 6, 2 / 3, 1 / 6],
            [1 / 6, 1 / 6, 2 / 3],
        ],
        dtype=float,
    )
    if not natural:
        p = np.array([n2l_tri(n, center=center) for n in nat], dtype=float)
    else:
        p = nat
    w = np.array([1 / 6, 1 / 6, 1 / 6])
    return p, w


def Gauss_Legendre_Tri_3b(
    center: Optional[Union[ndarray, None]] = None, natural: Optional[bool] = False
) -> Tuple[ndarray, ndarray]:
    nat = np.array(
        [
            [0.0, 1 / 2, 1 / 2],
            [1 / 2, 0.0, 1 / 2],
            [1 / 2, 1 / 2, 0.0],
        ],
        dtype=float,
    )
    if not natural:
        p = np.array([n2l_tri(n, center=center) for n in nat], dtype=float)
    else:
        p = nat
    w = np.array([1 / 6, 1 / 6, 1 / 6])
    return p, w


def Gauss_Legendre_Tri_4(
    center: Optional[Union[ndarray, None]] = None, natural: Optional[bool] = False
) -> Tuple[ndarray, ndarray]:
    nat = np.array(
        [
            [1 / 3, 1 / 3, 1 / 3],
            [0.2, 0.6, 0.2],
            [0.2, 0.2, 0.6],
            [0.6, 0.2, 0.2],
        ],
        dtype=float,
    )
    if not natural:
        p = np.array([n2l_tri(n, center=center) for n in nat], dtype=float)
    else:
        p = nat
    w = np.array([-0.5625, 0.520833333333333, 0.520833333333333, 0.520833333333333]) / 2
    return p, w


def Gauss_Legendre_Tri_6(
    center: Optional[Union[ndarray, None]] = None, natural: Optional[bool] = False
) -> Tuple[ndarray, ndarray]:
    nat = np.array(
        [
            [0.445948490915965, 0.108103018168070],
            [0.445948490915965, 0.445948490915965],
            [0.108103018168070, 0.445948490915965],
            [0.091576213509771, 0.816847572980459],
            [0.091576213509771, 0.091576213509771],
            [0.816847572980459, 0.091576213509771],
        ],
        dtype=float,
    )
    nat = _complete_natural_coordinates(nat)
    if not natural:
        p = np.array([n2l_tri(n, center=center) for n in nat], dtype=float)
    else:
        p = nat
    w = (
        np.array(
            [
                0.223381589678011,
                0.223381589678011,
                0.223381589678011,
                0.109951743655322,
                0.109951743655322,
                0.109951743655322,
            ]
        )
        / 2
    )
    return p, w


#  QUADRILATERALS


def Gauss_Legendre_Quad_Grid(i: int, j: int = None) -> Tuple[ndarray, ndarray]:
    j = i if j is None else j
    return gp(i, j)


def Gauss_Legendre_Quad_1() -> Tuple[ndarray, ndarray]:
    return gp(1, 1)


def Gauss_Legendre_Quad_4() -> Tuple[ndarray, ndarray]:
    return gp(2, 2)


def Gauss_Legendre_Quad_9() -> Tuple[ndarray, ndarray]:
    return gp(3, 3)


#  TETRAHEDRA


def Gauss_Legendre_Tet_1(
    center: Optional[Union[ndarray, None]] = None, natural: Optional[bool] = False
) -> Tuple[ndarray, ndarray]:
    nat = np.array([[0.25, 0.25, 0.25, 0.25]])
    if not natural:
        p = np.array([n2l_tet(n, center=center) for n in nat], dtype=float)
    else:
        p = nat
    w = np.array([1 / 6])
    return p, w


def Gauss_Legendre_Tet_4(
    center: Optional[Union[ndarray, None]] = None, natural: Optional[bool] = False
) -> Tuple[ndarray, ndarray]:
    nat = np.array(
        [
            [
                0.585410196624968,
                0.138196601125010,
                0.138196601125010,
                0.138196601125010,
            ],
            [
                0.138196601125010,
                0.585410196624968,
                0.138196601125010,
                0.138196601125010,
            ],
            [
                0.138196601125010,
                0.138196601125010,
                0.585410196624968,
                0.138196601125010,
            ],
            [
                0.138196601125010,
                0.138196601125010,
                0.138196601125010,
                0.585410196624968,
            ],
        ]
    )
    if not natural:
        p = np.array([n2l_tet(n, center=center) for n in nat], dtype=float)
    else:
        p = nat
    w = np.full(4, 1 / 24)
    return p, w


def Gauss_Legendre_Tet_5(
    center: Optional[Union[ndarray, None]] = None, natural: Optional[bool] = False
) -> Tuple[ndarray, ndarray]:
    nat = np.array(
        [
            [1 / 4, 1 / 4, 1 / 4, 1 / 4],
            [1 / 2, 1 / 6, 1 / 6, 1 / 6],
            [1 / 6, 1 / 2, 1 / 6, 1 / 6],
            [1 / 6, 1 / 6, 1 / 2, 1 / 6],
            [1 / 6, 1 / 6, 1 / 6, 1 / 2],
        ]
    )
    if not natural:
        p = np.array([n2l_tet(n, center=center) for n in nat], dtype=float)
    else:
        p = nat
    w = np.array([-4 / 30, 9 / 120, 9 / 120, 9 / 120, 9 / 120])
    return p, w


def Gauss_Legendre_Tet_11(
    center: Optional[Union[ndarray, None]] = None, natural: Optional[bool] = False
) -> Tuple[ndarray, ndarray]:
    nat = np.array(
        [
            [1 / 4, 1 / 4, 1 / 4, 1 / 4],
            [
                0.785714285714286,
                0.0714285714285714,
                0.0714285714285714,
                0.0714285714285714,
            ],
            [
                0.0714285714285714,
                0.785714285714286,
                0.0714285714285714,
                0.0714285714285714,
            ],
            [
                0.0714285714285714,
                0.0714285714285714,
                0.785714285714286,
                0.0714285714285714,
            ],
            [
                0.0714285714285714,
                0.0714285714285714,
                0.0714285714285714,
                0.785714285714286,
            ],
            [
                0.399403576166799,
                0.399403576166799,
                0.100596423833201,
                0.100596423833201,
            ],
            [
                0.399403576166799,
                0.100596423833201,
                0.399403576166799,
                0.100596423833201,
            ],
            [
                0.399403576166799,
                0.100596423833201,
                0.100596423833201,
                0.399403576166799,
            ],
            [
                0.100596423833201,
                0.399403576166799,
                0.399403576166799,
                0.100596423833201,
            ],
            [
                0.100596423833201,
                0.399403576166799,
                0.100596423833201,
                0.399403576166799,
            ],
            [
                0.100596423833201,
                0.100596423833201,
                0.399403576166799,
                0.399403576166799,
            ],
        ]
    )
    if not natural:
        p = np.array([n2l_tet(n, center=center) for n in nat], dtype=float)
    else:
        p = nat
    w = np.array(
        [
            -74 / 5625,
            343 / 45000,
            343 / 45000,
            343 / 45000,
            343 / 45000,
            56 / 2250,
            56 / 2250,
            56 / 2250,
            56 / 2250,
            56 / 2250,
            56 / 2250,
        ]
    )
    return p, w


#  HEXAHEDRA


def Gauss_Legendre_Hex_Grid(
    i: int, j: int = None, k: int = None
) -> Tuple[ndarray, ndarray]:
    j = i if j is None else j
    k = j if k is None else k
    return gp(i, j, k)


# WEDGES


def Gauss_Legendre_Wedge_3x2() -> Tuple[ndarray, ndarray]:
    p_tri, w_tri = Gauss_Legendre_Tri_3a()
    p_line, w_line = Gauss_Legendre_Line_Grid(2)
    p = np.zeros((6, 3), dtype=float)
    w = np.zeros((6,), dtype=float)
    p[:3, :2] = p_tri
    p[:3, 2] = p_line[0]
    w[:3] = w_tri * w_line[0]
    p[3:6, :2] = p_tri
    p[3:6, 2] = p_line[0]
    w[3:6] = w_tri * w_line[1]
    return p, w


def Gauss_Legendre_Wedge_3x3() -> Tuple[ndarray, ndarray]:
    p_tri, w_tri = Gauss_Legendre_Tri_3a()
    p_line, w_line = Gauss_Legendre_Line_Grid(3)
    n = len(w_line) * len(w_tri)
    p = np.zeros((n, 3), dtype=float)
    w = np.zeros((n,), dtype=float)
    for i in range(len(w_line)):
        p[i * 3 : (i + 1) * 3, :2] = p_tri
        p[i * 3 : (i + 1) * 3, 2] = p_line[i]
        w[i * 3 : (i + 1) * 3] = w_tri * w_line[i]
    return p, w



from ..config import __haspyvista__
from ..helpers import importers

if not __haspyvista__:  # pragma: no cover

    def from_pv(*_) -> None:
        raise ImportError(
            "You need PyVista for this. Install it with 'pip install pyvista'. "
            "You may also need to restart your kernel and reload the package."
        )

else:
    import pyvista as pv
    from typing import Union

    from ..data import PolyData, PointData, PolyCell
    from ..space import CartesianFrame

    from ..helpers import vtk_to_celltype
    from ..utils.space import frames_of_surfaces, frames_of_lines

    pyVistaLike = Union[pv.PolyData, pv.PointGrid, pv.UnstructuredGrid]

    def from_pv(pvobj: pyVistaLike) -> PolyData:
        """
        Returns a :class:`~sigmaepsilon.mesh.polydata.PolyData` instance from
        a :class:`pyvista.PolyData` or a :class:`pyvista.UnstructuredGrid`
        instance.

        .. note::
            See https://github.com/pyvista/pyvista for more examples with
            ``pyvista``. Be sure to install ``pyvista`` with ``pip install
            pyvista`` if you wish to use it.

        Example
        -------
        >>> from pyvista import examples
        >>> from sigmaepsilon.mesh import PolyData
        >>> bunny = examples.download_bunny_coarse()
        >>> mesh = PolyData.from_pv(bunny)
        """
        coords, cells_dict = None, None

        if isinstance(pvobj, pv.UnstructuredGrid):
            coords = pvobj.points.astype(float)
            cells_dict = pvobj.cells_dict
        else:
            try:
                ugrid = pvobj.cast_to_unstructured_grid()
                return from_pv(ugrid)
            except Exception:
                raise TypeError(f"Can't import from type {type(pvobj)}.")

        GlobalFrame = CartesianFrame(dim=3)
        pd = PointData(coords=coords, frame=GlobalFrame)
        polydata = PolyData(pd)

        for vtkid, vtktopo in cells_dict.items():
            if vtkid in vtk_to_celltype:
                celltype: PolyCell = vtk_to_celltype[vtkid]

                NDIM = celltype.Geometry.number_of_spatial_dimensions
                if NDIM == 1:
                    frames = frames_of_lines(coords, vtktopo)
                elif NDIM == 2:
                    frames = frames_of_surfaces(coords, vtktopo)
                elif NDIM == 3:
                    frames = GlobalFrame

                cd = celltype(topo=vtktopo, frames=frames)
                polydata[vtkid] = PolyData(cd)
            else:  # pragma: no cover
                raise NotImplementedError(
                    f"The element type with vtkId <{vtkid}> is not yet supported here."
                )

        return polydata


importers["PyVista"] = from_pv

__all__ = ["from_pv"]



from ...config import __hasplotly__

if not __hasplotly__:  # pragma: no cover

    def scatter_points_plotly(*_, **__):
        raise ImportError(
            "You need Plotly for this. Install it with 'pip install plotly'. "
            "You may also need to restart your kernel and reload the package."
        )

else:
    from typing import Iterable, Optional, Union
    from numbers import Number

    import plotly.express as px
    import plotly.graph_objects as go
    import pandas as pd
    import numpy as np
    from numpy import ndarray

    from sigmaepsilon.math import atleast2d

    def scatter_points_plotly(
        coords: Optional[ndarray],
        *,
        scalars: Optional[Union[ndarray, None]] = None,
        markersize: Optional[Number] = 1,
        scalar_labels: Optional[Union[Iterable[str], None]] = None,
    ) -> go.Figure:
        """
        Convenience function to plot several points in 3d with data and labels.

        Parameters
        ----------
        coords: numpy.ndarray
            The coordinates of the points, where the first axis runs along the points, the
            second along spatial dimensions.
        scalars: numpy.ndarray
            The values to show in the tooltips of the points as a 1d or 2d NumPy array.
            The length of the array must equal the number of points. Default is None.
        markersize: int, Optional
            The size of the balls at the point coordinates. Default is 1.
        scalar_labels: Iterable[str], Optional
            The labels for the columns in 'scalars'. Default is None.

        Example
        -------
        .. plotly::
            :include-source: True

            from sigmaepsilon.mesh.plotting import scatter_points_plotly
            import numpy as np
            points = np.array([
                [0, 0, 0],
                [1, 0, 0],
                [0, 1, 0],
                [0, 0, 1]
            ])
            data = np.random.rand(len(points))
            scalar_labels=["random data"]
            scatter_points_plotly(points, scalars=data, scalar_labels=scalar_labels)
        """
        x = coords[:, 0]
        y = coords[:, 1]
        z = coords[:, 2]

        dfdata = {
            "x": x,
            "y": y,
            "z": z,
            "size": np.full(len(x), markersize),
            "symbol": np.full(len(x), 4),
            "id": np.arange(1, len(x) + 1),
        }

        if scalars is not None:
            scalars = atleast2d(scalars, back=True)

        if scalars is not None and scalar_labels is None:
            nData = scalars.shape[1]
            scalar_labels = ["#{}".format(i + 1) for i in range(nData)]

        if scalar_labels is not None:
            assert len(scalar_labels) == scalars.shape[1]
            hover_data = scalar_labels
            sdata = {scalar_labels[i]: scalars[:, i] for i in range(len(scalar_labels))}
            dfdata.update(sdata)
        else:
            hover_data = ["x", "y", "z"]

        custom_data = scalar_labels if scalars is not None else None

        df = pd.DataFrame.from_dict(dfdata)

        fig = px.scatter_3d(
            df,
            x="x",
            y="y",
            z="z",
            hover_name="id",
            hover_data=hover_data,
            size="size",
            text="id",
            custom_data=custom_data,
        )

        if scalars is not None:
            tmpl = lambda i: "{" + "customdata[{}]:.4e".format(i) + "}"
            lbl = lambda i: scalar_labels[i]
            fnc = lambda i: "{label}: %{index}".format(label=lbl(i), index=tmpl(i))
            labels = [fnc(i) for i in range(len(scalar_labels))]
            fig.update_traces(hovertemplate="<br>".join(labels))
        else:
            fig.update_traces(
                hovertemplate="<br>".join(
                    [
                        "x: %{x}",
                        "y: %{y}",
                        "z: %{z}",
                    ]
                )
            )

        return fig


__all__ = ["scatter_points_plotly"]



from typing import Tuple, Optional, Any

import numpy as np
from numpy import ndarray

from sigmaepsilon.math import ascont
from sigmaepsilon.math.linalg import ReferenceFrame

from .polydata import PolyData
from .pointdata import PointData
from ..cells import TET4
from ..utils.space import frames_of_surfaces, is_planar_surface as is_planar
from ..extrude import extrude_T3_TET4
from ..triang import triangulate
from ..utils.tri import edges_tri
from ..utils.topology import unique_topo_data, T6_to_T3


__all__ = ["TriMesh"]


class TriMesh(PolyData):
    """
    A class to handle triangular meshes.

    All positional and keyword arguments not listed here are forwarded to
    :class:`~sigmaepsilon.mesh.data.polydata.PolyData`.

    Besides the arguments `coords` and `topo`, this class accepts these
    using aliases `points` and `triangles`.

    Parameters
    ----------
    points: numpy.ndarray, Optional
        2d numpy array of floats, describing a pointcloud. Default is None.
    triangles: numpy.ndarray, Optional
        2d numpy array of integers, describing the topology of a polygonal mesh.
        Default is None.

    Notes
    -----
    See the PolyData class for the rest of the possible arguments to the
    creator of this class. Note that, `points` and `triangles` are aliases
    to `coords` and `topo`, but the original terminology is still available.

    Examples
    --------
    Triangulate a rectangle of size 800x600 with a subdivision of 10x10
    and calculate the area

    >>> from sigmaepsilon.mesh import TriMesh, CartesianFrame, PointData, triangulate
    >>> from sigmaepsilon.mesh.cells import T3
    >>> import numpy as np
    >>> frame = CartesianFrame(dim=3)
    >>> coords, topo, _ = triangulate(size=(800, 600), shape=(10, 10))
    >>> pd = PointData(coords=coords, frame=frame)
    >>> cd = T3(topo=topo)
    >>> trimesh = TriMesh(pd, cd)
    >>> np.isclose(trimesh.area(), 480000.0)
    True

    Extrude to create a tetrahedral mesh

    >>> tetmesh = trimesh.extrude(h=300, N=5)
    >>> np.isclose(tetmesh.volume(), 144000000.0)
    True

    Calculate normals and tell if the triangles form
    a planar surface or not

    >>> normals = trimesh.normals()
    >>> trimesh.is_planar()
    True

    Create a circular disk

    >>> from sigmaepsilon.mesh.recipes import circular_disk
    >>> trimesh = circular_disk(120, 60, 5, 25)

    See Also
    --------
    :class:`~sigmaepsilon.mesh.data.polydata.PolyData`
    :class:`~sigmaepsilon.mesh.space.frame.CartesianFrame`
    """

    def axes(self) -> np.ndarray:
        """
        Returns the normalized coordinate frames of triangles as a 3d numpy array.
        """
        x = self.coords()
        assert x.shape[-1] == 3, "This is only available for 3d datasets."
        return frames_of_surfaces(x, self.topology().to_numpy()[:, :3])

    def normals(self) -> np.ndarray:
        """
        Retuns the surface normals as a 2d numpy array.
        """
        return ascont(self.axes()[:, 2, :])

    def is_planar(self) -> bool:
        """
        Returns `True` if the triangles form a planar surface.
        """
        return is_planar(self.normals())

    def extrude(self, *, h: float, N: int) -> PolyData:
        """
        Exctrude mesh perpendicular to the plane of the triangulation.
        The target element type can be specified with the `celltype` argument.

        Parameters
        ----------
        h: float
            Size perpendicular to the plane of the surface to be extruded.
        N: int
            Number of subdivisions along the perpendicular direction.

        Returns
        -------
        :class:`~sigmaepsilon.mesh.tetmesh.TetMesh`
            A tetrahedral mesh.
        """
        if not self.is_planar():
            raise RuntimeError("Only planar surfaces can be extruded!")

        frame = ReferenceFrame(self.cd.frames[0])
        x = self.coords(target=frame)[:, :2]
        x, topo = extrude_T3_TET4(x, self.topology().to_numpy()[:, :3], h, N)
        pd = PointData(coords=x, frame=frame)
        cd = TET4(topo=topo, frames=frame)
        return PolyData(pd, cd)

    def edges(self, return_cells: bool = False) -> Tuple[ndarray, Optional[ndarray]]:
        """
        Returns point indices of the unique edges in the model.
        If `return_cells` is `True`, it also returns the edge
        indices of the triangles, referencing the edges.

        Parameters
        ----------
        return_cells: bool, Optional
            If True, returns the edge indices of the triangles,
            that can be used to reconstruct the topology.
            Default is False.

        Returns
        -------
        numpy.ndarray
            Integer array of indices, representing point indices of edges.

        numpy.ndarray, Optional
            Integer array of indices, that together with the edge data
            reconstructs the topology.
        """
        edges, IDs = unique_topo_data(edges_tri(self.topology().to_numpy()))
        if return_cells:
            return edges, IDs
        else:
            return edges

    def to_triobj(self) -> Any:
        """
        Returns a triangulation object of a specified backend.
        See :func:`~sigmaepsilon.mesh.triang.triangulate` for the details.

        Note
        ----
        During the process, the 6-noded triangles of the section are converted
        into 3-noded ones.

        See also
        --------
        :class:`~matplotlib.tri.Triangulation`
        :func:`~sigmaepsilon.mesh.triang.triangulate`
        """
        coords, topo = self.coords(), self.topology().to_numpy()
        if topo.shape[-1] == 6:
            path = np.array([[0, 5, 4], [5, 1, 3], [3, 2, 4], [5, 3, 4]], dtype=int)
            coords, topo = T6_to_T3(coords, topo, path=path)
        return triangulate(points=coords, triangles=topo)[-1]



from .geometry import GeometryProtocol
from .abcpolycell import ABC_PolyCell
from .data import (
    PointDataProtocol,
    CellDataProtocol,
    PolyCellProtocol,
    PolyDataProtocol,
    PolyDataLike,
    PointDataLike,
    CellDataLike,
    PolyCellLike,
)

__all__ = [
    "PointDataProtocol",
    "PolyDataProtocol",
    "CellDataProtocol",
    "GeometryProtocol",
    "PolyCellProtocol",
    "ABC_PolyCell",
    "PolyDataLike",
    "PointDataLike",
    "CellDataLike",
    "PolyCellLike",
]



# -*- coding: utf-8 -*-
from .frame import CartesianFrame
from .frame import CartesianFrame as StandardFrame
from .point import Point
from .pointcloud import PointCloud


__all__ = ["CartesianFrame", "StandardFrame", "Point", "PointCloud"]



from typing import Tuple, List
from functools import partial

from sympy import symbols
import numpy as np
from numpy import ndarray

from ..geometry import PolyCellGeometry3d
from ..data.polycell import PolyCell
from ..utils.cells.h8 import (
    shp_H8_multi,
    dshp_H8_multi,
    shape_function_matrix_H8_multi,
    monoms_H8,
)
from ..utils.numint import Gauss_Legendre_Hex_Grid


class H8(PolyCell):
    """
    8-node hexahedron.

    ::

        top
        7--6
        |  |
        4--5

        bottom
        3--2
        |  |
        0--1
    """

    label = "H8"

    class Geometry(PolyCellGeometry3d):
        number_of_nodes = 8
        vtk_cell_id = 12
        shape_function_evaluator = shp_H8_multi
        shape_function_matrix_evaluator = shape_function_matrix_H8_multi
        shape_function_derivative_evaluator = dshp_H8_multi
        monomial_evaluator = monoms_H8
        quadrature = {
            "full": partial(Gauss_Legendre_Hex_Grid, 2, 2, 2),
            "geometry": "full",
        }

        @classmethod
        def polybase(cls) -> Tuple[List]:
            """
            Retruns the polynomial base of the master element.

            Returns
            -------
            list
                A list of SymPy symbols.
            list
                A list of monomials.
            """
            locvars = r, s, t = symbols("r s t", real=True)
            monoms = [1, r, s, t, r * s, r * t, s * t, r * s * t]
            return locvars, monoms

        @classmethod
        def master_coordinates(cls) -> ndarray:
            """
            Returns local coordinates of the cell.

            Returns
            -------
            numpy.ndarray
            """
            return np.array(
                [
                    [-1.0, -1.0, -1],
                    [1.0, -1.0, -1.0],
                    [1.0, 1.0, -1.0],
                    [-1.0, 1.0, -1.0],
                    [-1.0, -1.0, 1.0],
                    [1.0, -1.0, 1.0],
                    [1.0, 1.0, 1.0],
                    [-1.0, 1.0, 1.0],
                ]
            )

        @classmethod
        def master_center(cls) -> ndarray:
            """
            Returns the local coordinates of the center of the cell.

            Returns
            -------
            numpy.ndarray
            """
            return np.array([0.0, 0.0, 0.0])

        @classmethod
        def tetmap(cls) -> np.ndarray:
            return np.array(
                [[1, 2, 0, 5], [3, 0, 2, 7], [5, 4, 7, 0], [6, 5, 7, 2], [0, 2, 7, 5]],
                dtype=int,
            )



from numba import njit, prange
import numpy as np
from numpy import ndarray

__cache = True


@njit(nogil=True, cache=__cache)
def monoms_T3(x: ndarray) -> ndarray:
    r, s = x
    return np.array([1, r, s], dtype=float)


@njit(nogil=True, cache=__cache)
def shp_T3(pcoord: ndarray) -> ndarray:
    r, s = pcoord
    return np.array([1 / 3 - r - s, r + 1 / 3, s + 1 / 3], dtype=pcoord.dtype)


@njit(nogil=True, parallel=True, cache=__cache)
def shp_T3_multi(pcoords: ndarray) -> ndarray:
    nP = pcoords.shape[0]
    res = np.zeros((nP, 3), dtype=pcoords.dtype)
    for iP in prange(nP):
        res[iP, :] = shp_T3(pcoords[iP])
    return res


@njit(nogil=True, parallel=False, cache=__cache)
def shape_function_matrix_T3(pcoord: ndarray, ndof: int = 2) -> ndarray:
    eye = np.eye(ndof, dtype=pcoord.dtype)
    shp = shp_T3(pcoord)
    res = np.zeros((ndof, ndof * 3), dtype=pcoord.dtype)
    for i in prange(3):
        res[:, i * ndof : (i + 1) * ndof] = eye * shp[i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def shape_function_matrix_T3_multi(pcoords: ndarray, ndof: int = 2) -> ndarray:
    nP = pcoords.shape[0]
    res = np.zeros((nP, ndof, ndof * 3), dtype=pcoords.dtype)
    for iP in prange(nP):
        res[iP] = shape_function_matrix_T3(pcoords[iP], ndof)
    return res


@njit(nogil=True, cache=__cache)
def dshp_T3(x) -> ndarray:
    return np.array([[-1.0, -1.0], [1.0, 0.0], [0.0, 1.0]])


@njit(nogil=True, parallel=True, cache=__cache)
def dshp_T3_multi(pcoords: ndarray) -> ndarray:
    nP = pcoords.shape[0]
    res = np.zeros((nP, 3, 2), dtype=pcoords.dtype)
    for iP in prange(nP):
        res[iP] = dshp_T3(pcoords[iP])
    return res



from .triplot import triplot_mpl_hinton, triplot_mpl_data, triplot_mpl_mesh
from .parallel import parallel_mpl, aligned_parallel_mpl
from .utils import decorate_mpl_ax

__all__ = [
    "triplot_mpl_hinton",
    "triplot_mpl_mesh",
    "triplot_mpl_data",
    "parallel_mpl",
    "aligned_parallel_mpl",
    "decorate_mpl_ax",
]



# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray as array
from numba import njit, prange

from sigmaepsilon.math import minmax

from .utils.topology import detach_mesh_bulk as detach_mesh, remap_topo

__cache = True


@njit(nogil=True, parallel=True, cache=__cache)
def tile2d(coords: array, topo: array, shape: tuple, tol: float = 1e-8):
    """
    Tiles a rectangle with the mesh defined by the coordinate
    array 'coords' and topology array 'topo'.

    Parameters
    ----------
    coords: ndarray

    topo: ndaray

    shape : Tupel

    tol : float

    """
    nX, nY = shape
    nP, nD = coords.shape
    nE, nNE = topo.shape

    # identify points on the boundaries
    xmin, xmax = minmax(coords[:, 0])
    ymin, ymax = minmax(coords[:, 1])
    dx = xmax - xmin
    dy = ymax - ymin
    inds_xmin = np.where(np.abs(coords[:, 0] - xmin) < tol)[0]
    inds_xmax = np.where(np.abs(coords[:, 0] - xmax) < tol)[0]
    inds_ymin = np.where(np.abs(coords[:, 1] - ymin) < tol)[0]
    inds_ymax = np.where(np.abs(coords[:, 1] - ymax) < tol)[0]

    # allocate space for the results
    nP_new = nX * nY * nP
    nE_new = nX * nY * nE
    coords_new = np.zeros((nP_new, nD), dtype=coords.dtype)
    topo_new = np.zeros((nE_new, nNE), dtype=topo.dtype)

    # tile mesh and prepare mapping to renumber coincident nodes
    imap = np.arange(nP_new, dtype=topo.dtype)
    for i in prange(nX):
        for j in prange(nY):
            blockId = i * nY + j
            i_start = blockId * nE
            i_stop = i_start + nE
            topo_new[i_start:i_stop, :] = topo + blockId * nP
            i_start = blockId * nP
            i_stop = i_start + nP
            coords_new[i_start:i_stop, 0] = coords[:, 0] + i * dx
            coords_new[i_start:i_stop, 1] = coords[:, 1] + j * dy
            coords_new[i_start:i_stop, 2] = coords[:, 2]
            # mapping
            if i > 0:
                inds_old = inds_xmin + blockId * nP
                inds_new = inds_xmax + ((i - 1) * nY + j) * nP
                imap[inds_old] = inds_new
            if j > 0:
                inds_old = inds_ymin + blockId * nP
                inds_new = inds_ymax + (i * nY + (j - 1)) * nP
                imap[inds_old] = inds_new

    # renumber and return detached mesh
    return detach_mesh(coords_new, remap_topo(topo_new, imap))



# -*- coding: utf-8 -*-
from typing import Tuple, List

import numpy as np
from numpy import ndarray
from sympy import symbols

from ..geometry import PolyCellGeometry2d
from ..data.polycell import PolyCell
from ..utils.numint import Gauss_Legendre_Tri_1
from ..utils.cells.t3 import (
    shp_T3_multi,
    dshp_T3_multi,
    shape_function_matrix_T3_multi,
    monoms_T3,
)
from ..utils.utils import points_of_cells
from ..utils.tri import area_tri_bulk


class T3(PolyCell):
    """
    A class to handle 3-noded triangles.

    Example
    -------
    >>> from sigmaepsilon.mesh import TriMesh, CartesianFrame, PointData, triangulate
    >>> from sigmaepsilon.mesh.cells import T3 as CellData
    >>> frame = CartesianFrame(dim=3)
    >>> coords, topo, _ = triangulate(size=(800, 600), shape=(10, 10))
    >>> pd = PointData(coords=coords, frame=frame)
    >>> cd = CellData(topo=topo)
    >>> trimesh = TriMesh(pd, cd)
    >>> np.isclose(trimesh.area(), 480000.0)
    True
    """

    label = "T3"

    class Geometry(PolyCellGeometry2d):
        number_of_nodes = 3
        vtk_cell_id = 5
        shape_function_evaluator = shp_T3_multi
        shape_function_matrix_evaluator = shape_function_matrix_T3_multi
        shape_function_derivative_evaluator = dshp_T3_multi
        monomial_evaluator = monoms_T3
        quadrature = {
            "full": Gauss_Legendre_Tri_1,
            "geometry": "full",
        }

        @classmethod
        def trimap(cls) -> ndarray:
            """
            Returns a mapping used to transform the topology to triangles.
            This is only implemented here for standardization.
            """
            return np.array([[0, 1, 2]], dtype=int)

        @classmethod
        def polybase(cls) -> Tuple[List]:
            """
            Retruns the polynomial base of the master element.

            Returns
            -------
            list
                A list of SymPy symbols.
            list
                A list of monomials.
            """
            locvars = r, s = symbols("r s", real=True)
            monoms = [1, r, s]
            return locvars, monoms

        @classmethod
        def master_coordinates(cls) -> ndarray:
            """
            Returns local coordinates of the master cell relative to the origo
            of the master cell.

            Returns
            -------
            numpy.ndarray
            """
            return np.array([[-1 / 3, -1 / 3], [2 / 3, -1 / 3], [-1 / 3, 2 / 3]])

        @classmethod
        def master_center(cls) -> ndarray:
            """
            Returns the center of the master cell relative to the origo
            of the master cell.

            Returns
            -------
            numpy.ndarray
            """
            return np.array([[0.0, 0.0]], dtype=float)

    def to_triangles(self) -> ndarray:
        """
        Returns the topology as triangles.
        """
        return self.topology().to_numpy()

    def areas(self, *_, **__) -> ndarray:
        """
        Returns the areas of the cells as an 1d NumPy array.
        """
        coords = self.container.source().coords()
        topo = self.topology().to_numpy()
        ec = points_of_cells(coords, topo, local_axes=self.frames)
        return area_tri_bulk(ec)

    @classmethod
    def from_TriMesh(
        cls, *args, coords: ndarray = None, topo: ndarray = None, **__
    ) -> Tuple[ndarray, ndarray]:
        from sigmaepsilon.mesh.data.trimesh import TriMesh

        if len(args) > 0 and isinstance(args[0], TriMesh):
            mesh = args[0]
            return mesh.coords(), mesh.topology().to_numpy()
        elif coords is not None and topo is not None:
            return coords, topo
        else:
            raise NotImplementedError



from sigmaepsilon.core.meta import ABCMeta_Weak


__all__ = ["ABCMeta_AkWrapper", "ABC_AkWrapper"]


class ABCMeta_AkWrapper(ABCMeta_Weak):
    """
    Meta class for PointData and CellData classes.

    It merges attribute maps with those of the parent classes.
    """

    def __init__(self, name, bases, namespace, *args, **kwargs):
        super().__init__(name, bases, namespace, *args, **kwargs)

    def __new__(metaclass, name, bases, namespace, *args, **kwargs):
        cls = super().__new__(metaclass, name, bases, namespace, *args, **kwargs)

        # merge database fields
        _attr_map_ = namespace.get("_attr_map_", {})
        for base in bases:
            _attr_map_.update(base.__dict__.get("_attr_map_", {}))
        cls._attr_map_ = _attr_map_

        return cls


class ABC_AkWrapper(metaclass=ABCMeta_AkWrapper):
    """
    Helper class that provides a standard way to create an ABC using
    inheritance.
    """

    __slots__ = ()
